@@ -14,11 +14,10 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from taskflow import exceptions
 from taskflow.patterns import graph_flow as gf
 from taskflow.patterns import linear_flow as lf
 from taskflow.patterns import unordered_flow as uf
-
-from taskflow import exceptions
 from taskflow import retry
 from taskflow import test
 from taskflow.tests import utils
@@ -28,40 +27,40 @@ class FlowDependenciesTest(test.TestCase):
 
     def test_task_without_dependencies(self):
         flow = utils.TaskNoRequiresNoReturns()
-        self.assertEqual(flow.requires, set())
-        self.assertEqual(flow.provides, set())
+        self.assertEqual(set(), flow.requires)
+        self.assertEqual(set(), flow.provides)
 
     def test_task_requires_default_values(self):
         flow = utils.TaskMultiArg()
-        self.assertEqual(flow.requires, set(['x', 'y', 'z']))
-        self.assertEqual(flow.provides, set())
+        self.assertEqual(set(['x', 'y', 'z']), flow.requires)
+        self.assertEqual(set(), flow.provides, )
 
     def test_task_requires_rebinded_mapped(self):
         flow = utils.TaskMultiArg(rebind={'x': 'a', 'y': 'b', 'z': 'c'})
-        self.assertEqual(flow.requires, set(['a', 'b', 'c']))
-        self.assertEqual(flow.provides, set())
+        self.assertEqual(set(['a', 'b', 'c']), flow.requires)
+        self.assertEqual(set(), flow.provides)
 
     def test_task_requires_additional_values(self):
         flow = utils.TaskMultiArg(requires=['a', 'b'])
-        self.assertEqual(flow.requires, set(['a', 'b', 'x', 'y', 'z']))
-        self.assertEqual(flow.provides, set())
+        self.assertEqual(set(['a', 'b', 'x', 'y', 'z']), flow.requires)
+        self.assertEqual(set(), flow.provides)
 
     def test_task_provides_values(self):
         flow = utils.TaskMultiReturn(provides=['a', 'b', 'c'])
-        self.assertEqual(flow.requires, set())
-        self.assertEqual(flow.provides, set(['a', 'b', 'c']))
+        self.assertEqual(set(), flow.requires)
+        self.assertEqual(set(['a', 'b', 'c']), flow.provides)
 
     def test_task_provides_and_requires_values(self):
         flow = utils.TaskMultiArgMultiReturn(provides=['a', 'b', 'c'])
-        self.assertEqual(flow.requires, set(['x', 'y', 'z']))
-        self.assertEqual(flow.provides, set(['a', 'b', 'c']))
+        self.assertEqual(set(['x', 'y', 'z']), flow.requires)
+        self.assertEqual(set(['a', 'b', 'c']), flow.provides)
 
     def test_linear_flow_without_dependencies(self):
         flow = lf.Flow('lf').add(
             utils.TaskNoRequiresNoReturns('task1'),
             utils.TaskNoRequiresNoReturns('task2'))
-        self.assertEqual(flow.requires, set())
-        self.assertEqual(flow.provides, set())
+        self.assertEqual(set(), flow.requires)
+        self.assertEqual(set(), flow.provides)
 
     def test_linear_flow_requires_values(self):
         flow = lf.Flow('lf').add(
@@ -131,24 +130,27 @@ class FlowDependenciesTest(test.TestCase):
 
     def test_unordered_flow_provides_required_values(self):
         flow = uf.Flow('uf')
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('task1', provides='x'),
-                          utils.TaskOneArg('task2'))
+        flow.add(utils.TaskOneReturn('task1', provides='x'),
+                 utils.TaskOneArg('task2'))
+        flow.add(utils.TaskOneReturn('task1', provides='x'),
+                 utils.TaskOneArg('task2'))
+        self.assertEqual(set(['x']), flow.provides)
+        self.assertEqual(set(['x']), flow.requires)
 
     def test_unordered_flow_requires_provided_value_other_call(self):
         flow = uf.Flow('uf')
         flow.add(utils.TaskOneReturn('task1', provides='x'))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneArg('task2'))
+        flow.add(utils.TaskOneArg('task2'))
+        self.assertEqual(set(['x']), flow.provides)
+        self.assertEqual(set(['x']), flow.requires)
 
     def test_unordered_flow_provides_required_value_other_call(self):
         flow = uf.Flow('uf')
         flow.add(utils.TaskOneArg('task2'))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('task1', provides='x'))
+        flow.add(utils.TaskOneReturn('task1', provides='x'))
+        self.assertEqual(2, len(flow))
+        self.assertEqual(set(['x']), flow.provides)
+        self.assertEqual(set(['x']), flow.requires)
 
     def test_unordered_flow_multi_provides_and_requires_values(self):
         flow = uf.Flow('uf').add(
@@ -162,16 +164,14 @@ class FlowDependenciesTest(test.TestCase):
 
     def test_unordered_flow_provides_same_values(self):
         flow = uf.Flow('uf').add(utils.TaskOneReturn(provides='x'))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn(provides='x'))
+        flow.add(utils.TaskOneReturn(provides='x'))
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_unordered_flow_provides_same_values_one_add(self):
         flow = uf.Flow('uf')
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn(provides='x'),
-                          utils.TaskOneReturn(provides='x'))
+        flow.add(utils.TaskOneReturn(provides='x'),
+                 utils.TaskOneReturn(provides='x'))
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_nested_flows_requirements(self):
         flow = uf.Flow('uf').add(
@@ -187,13 +187,6 @@ class FlowDependenciesTest(test.TestCase):
         self.assertEqual(flow.requires, set(['a', 'b', 'c']))
         self.assertEqual(flow.provides, set(['x', 'y', 'z', 'q']))
 
-    def test_nested_flows_provides_same_values(self):
-        flow = lf.Flow('lf').add(
-            uf.Flow('uf').add(utils.TaskOneReturn(provides='x')))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          gf.Flow('gf').add(utils.TaskOneReturn(provides='x')))
-
     def test_graph_flow_requires_values(self):
         flow = gf.Flow('gf').add(
             utils.TaskOneArg('task1'),
@@ -225,9 +218,8 @@ class FlowDependenciesTest(test.TestCase):
     def test_graph_flow_provides_provided_value_other_call(self):
         flow = gf.Flow('gf')
         flow.add(utils.TaskOneReturn('task1', provides='x'))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('task2', provides='x'))
+        flow.add(utils.TaskOneReturn('task2', provides='x'))
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_graph_flow_multi_provides_and_requires_values(self):
         flow = gf.Flow('gf').add(
@@ -251,10 +243,10 @@ class FlowDependenciesTest(test.TestCase):
                                                           requires=['a']))
 
     def test_task_requires_and_provides_same_values(self):
-        self.assertRaises(exceptions.DependencyFailure,
-                          utils.TaskOneArgOneReturn,
-                          requires='a',
-                          provides='a')
+        flow = lf.Flow('lf', utils.TaskOneArgOneReturn('rt', requires='x',
+                                                       provides='x'))
+        self.assertEqual(flow.requires, set('x'))
+        self.assertEqual(flow.provides, set('x'))
 
     def test_retry_in_linear_flow_no_requirements_no_provides(self):
         flow = lf.Flow('lf', retry.AlwaysRevert('rt'))
@@ -279,9 +271,11 @@ class FlowDependenciesTest(test.TestCase):
         self.assertEqual(flow.provides, set(['a', 'b']))
 
     def test_retry_requires_and_provides_same_value(self):
-        self.assertRaises(exceptions.DependencyFailure,
-                          retry.AlwaysRevert,
-                          'rt', requires=['x', 'y'], provides=['x', 'y'])
+        flow = lf.Flow('lf', retry.AlwaysRevert('rt',
+                                                requires=['x', 'y'],
+                                                provides=['x', 'y']))
+        self.assertEqual(flow.requires, set(['x', 'y']))
+        self.assertEqual(flow.provides, set(['x', 'y']))
 
     def test_retry_in_unordered_flow_no_requirements_no_provides(self):
         flow = uf.Flow('uf', retry.AlwaysRevert('rt'))
@@ -337,18 +331,6 @@ class FlowDependenciesTest(test.TestCase):
         self.assertEqual(flow.requires, set(['x', 'y', 'c']))
         self.assertEqual(flow.provides, set(['a', 'b', 'z']))
 
-    def test_linear_flow_retry_and_task_dependency_conflict(self):
-        flow = lf.Flow('lf', retry.AlwaysRevert('rt', requires=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn(provides=['x']))
-
-    def test_linear_flow_retry_and_task_provide_same_value(self):
-        flow = lf.Flow('lf', retry.AlwaysRevert('rt', provides=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('t1', provides=['x']))
-
     def test_unordered_flow_retry_and_task(self):
         flow = uf.Flow('uf', retry.AlwaysRevert('rt',
                                                 requires=['x', 'y'],
@@ -359,24 +341,22 @@ class FlowDependenciesTest(test.TestCase):
         self.assertEqual(flow.requires, set(['x', 'y', 'c']))
         self.assertEqual(flow.provides, set(['a', 'b', 'z']))
 
-    def test_unordered_flow_retry_and_task_dependency_conflict(self):
+    def test_unordered_flow_retry_and_task_same_requires_provides(self):
         flow = uf.Flow('uf', retry.AlwaysRevert('rt', requires=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn(provides=['x']))
+        flow.add(utils.TaskOneReturn(provides=['x']))
+        self.assertEqual(set(['x']), flow.requires)
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_unordered_flow_retry_and_task_provide_same_value(self):
         flow = uf.Flow('uf', retry.AlwaysRevert('rt', provides=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('t1', provides=['x']))
+        flow.add(utils.TaskOneReturn('t1', provides=['x']))
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_unordered_flow_retry_two_tasks_provide_same_value(self):
         flow = uf.Flow('uf', retry.AlwaysRevert('rt', provides=['y']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('t1', provides=['x']),
-                          utils.TaskOneReturn('t2', provides=['x']))
+        flow.add(utils.TaskOneReturn('t1', provides=['x']),
+                 utils.TaskOneReturn('t2', provides=['x']))
+        self.assertEqual(set(['x', 'y']), flow.provides)
 
     def test_graph_flow_retry_and_task(self):
         flow = gf.Flow('gf', retry.AlwaysRevert('rt',
@@ -388,32 +368,16 @@ class FlowDependenciesTest(test.TestCase):
         self.assertEqual(flow.requires, set(['x', 'y', 'c']))
         self.assertEqual(flow.provides, set(['a', 'b', 'z']))
 
-    def test_graph_flow_retry_and_task_dependency_conflict(self):
+    def test_graph_flow_retry_and_task_dependency_provide_require(self):
         flow = gf.Flow('gf', retry.AlwaysRevert('rt', requires=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn(provides=['x']))
+        flow.add(utils.TaskOneReturn(provides=['x']))
+        self.assertEqual(set(['x']), flow.provides)
+        self.assertEqual(set(['x']), flow.requires)
 
     def test_graph_flow_retry_and_task_provide_same_value(self):
         flow = gf.Flow('gf', retry.AlwaysRevert('rt', provides=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          utils.TaskOneReturn('t1', provides=['x']))
-
-    def test_two_retries_provide_same_values_in_nested_flows(self):
-        flow = lf.Flow('lf', retry.AlwaysRevert('rt1', provides=['x']))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          lf.Flow('lf1', retry.AlwaysRevert('rt2',
-                                                            provides=['x'])))
-
-    def test_two_retries_provide_same_values(self):
-        flow = lf.Flow('lf').add(
-            lf.Flow('lf1', retry.AlwaysRevert('rt1', provides=['x'])))
-        self.assertRaises(exceptions.DependencyFailure,
-                          flow.add,
-                          lf.Flow('lf2', retry.AlwaysRevert('rt2',
-                                                            provides=['x'])))
+        flow.add(utils.TaskOneReturn('t1', provides=['x']))
+        self.assertEqual(set(['x']), flow.provides)
 
     def test_builtin_retry_args(self):
 
